<?php
/**
 * AstroPay Payment Gateway
 *
 * @description Provides AstroPay Payment Gateway for WooCommerce.
 *
 * @class WC_Gateway_AstroPay
 * @package WooCommerce
 * @category Payment Gateways
 * @author Gaurav Agrawal
 */

class WC_Gateway_AstroPay extends WC_Payment_Gateway {

	/**
	 * List of error messages corresponding to error codes
	 * 
	 * @since 1.0.0
	 */
	protected const ERRORS = [
		400 => 'Incorrect request body format.',
		401 => 'Incorrect API-Key',
		404 => 'There is no cashout with that reference.',
		404 => 'Missing configuration, please contact support.',
		404 => 'There is no deposit with that reference.',
		404 => 'There is no user with that reference.',
		412 => 'Cashout with this ID already exists.',
		412 => 'A Deposit with this ID have already been approved.',
		412 => 'Deposit with this ID already exists.',
		412 => 'Deposit with this ID has been cancelled.',
		412 => 'There is not enough balance to afford this transaction.',
		412 => 'Astropay Card is invalid.',
		412 => 'Invalid cashout. For further info contact support.',
		412 => 'Invalid configuration. For further info contact support.',
		412 => 'Invalid content type, should be appliction/Json.',
		412 => 'User located in Invalid Region.',
		412 => 'Signature was nos calculated correctly.',
		500 => 'Internal error, please contact support.',		
	];

	/**
	 * List of payment method names corresponding to payment method codes
	 * 
	 * @since 1.0.0
	 */
	protected const PAYMENT_METHODS = [
		'AB' => 'Agribank',
		'AE' => 'American Express',
		'AL' => 'Airtel Money',
		'AS' => 'Asia Commercial Bank',
		'BB' => 'Banco Brasil',
		'BC' => 'BCP',
		'BD' => 'BIDV Bank',
		'BL' => 'Boleto',
		'BQ' => 'Banorte',
		'BT' => 'Bitcoin',
		'BU' => 'Baloto',
		'BV' => 'Bancomer',
		'BW' => 'Bodega Aurrera',
		'BX' => 'Banco de Chile',
		'BY' => 'CIMB Bank Berhad',
		'CR' => 'Carulla',
		'CU' => 'Circulo K',
		'CZ' => 'Bank Central Asia',
		'DC' => 'Diners',
		'DO' => 'DongA Bank',
		'EF' => 'PagoEfectivo',
		'EL' => 'Elo',
		'EM' => 'Eximbank',
		'EX' => 'Almacenes Exito',
		'EY' => 'Efecty',
		'FA' => 'Farmacias del ahorro',
		'FB' => 'Farmacia Benavides',
		'FC' => 'FreeCharge',
		'FO' => 'Facilito',
		'GO' => 'Google Pay',
		'GS' => 'Government Savings Bank',
		'HC' => 'Caja Huancayo',
		'HO' => 'Hong Leong Bank Berhad',
		'HP' => 'Hipercard',
		'IA' => 'Itau',
		'IB' => 'InterBank',
		'IG' => 'Internet Banking',
		'IR' => 'Interac Online',
		'IX' => 'Pix',
		'JM' => 'Jio Money',
		'KB' => 'Bangkok Bank',
		'KR' => 'Krungsri',
		'KS' => 'Kasikorn Bank',
		'KT' => 'Krung Thai Bank',
		'LC' => 'Loterias Caixa',
		'MB' => 'Mandiri Bank',
		'MC' => 'Mastercard',
		'MJ' => 'Multicaja',
		'MM' => 'Mobile Money',
		'MP' => 'M-Pesa',
		'MY' => 'Maybank Berhad',
		'NB' => 'Net Banking India',
		'NG' => 'Bank Negara Indonesia',
		'OM' => 'Ola Money',
		'OX' => 'OXXO',
		'PC' => 'PSE',
		'PH' => 'PhonePe',
		'PM' => 'Perfect Money',
		'PU' => 'Public Bank Berhad',
		'RH' => 'RHB Banking Group',
		'SA' => 'Siam Commercial Bank',
		'SC' => 'Santander',
		'SE' => 'Spei',
		'SF' => 'Banco Safra',
		'SJ' => 'Banco Sicredi',
		'SK' => 'Sacombank',
		'SS' => 'Sams Club',
		'SU' => 'Superama',
		'SX' => 'Surtimax',
		'TB' => 'Thai Military Bank',
		'TC' => 'ToditoCash',
		'TH' => 'Techcombank',
		'TL' => 'Trustly',
		'TR' => 'Bank Transfer',
		'UD' => 'Ussd',
		'UI' => 'UPI',
		'UL' => 'Banrisul',
		'US' => 'Caja Cusco',
		'VC' => 'Virtual Account - Bank Negara Indonesia',
		'VE' => 'Verve',
		'VI' => 'Visa',
		'VJ' => 'Virtual Account - Mandiri Bank Indonesia',
		'VM' => 'Virtual Account - MayBank Indonesia',
		'VN' => 'Vietin Bank',
		'VT' => 'Vietcombank',
		'WA' => 'Walmart',
		'WP' => 'WebPay',
		'WU' => 'Western Union',
		'ZB' => 'Zimpler Banking',
		'ZP' => 'Zimpler'
	];


	/**
	 * __construct function.
	 *
	 * @since 1.0.0
	 * @access public
	 * @return void
	 */
	public function __construct() {
		global $woocommerce;

		$this->id				= 'astropay';

        $this->method_title 	= 'AstroPay';
		$this->method_description = __( 'Accept payments using AstroPay.', 'astropay' );
        $this->icon 			= plugins_url( 'assets/images/logo.svg', __DIR__ );
        $this->has_fields 		= false;
        $this->liveurl			= 'https://onetouch-api.astropay.com/';
        $this->testurl			= 'https://onetouch-api-sandbox.astropay.com/';
		$this->currency    		= get_woocommerce_currency(); //$this->get_currency();
		$this->country			= WC()->countries->get_base_country();
		$this->supports         = [ 'products', 'refunds' ];

      	// Load the form fields
		$this->init_form_fields();

		// Load the settings ( built-in )
		$this->init_settings();

		// Create a logger
		$this->logger			= class_exists( 'WC_Logger' ) ? new WC_Logger() : $woocommerce->logger();

		// Load parameter from settings
		$this->enabled			= $this->settings['enabled'];
		$this->title 			= $this->settings['title'];
		$this->description  	= $this->settings['description'];
		$this->x_login   		= $this->settings['x_login'];
		$this->secret   		= $this->settings['secret'];
		$this->testmode			= filter_var( $this->settings['testmode'], FILTER_VALIDATE_BOOLEAN );
		$this->merchant_name	= $this->settings['merchant_name'];
		$this->merchant_logo	= $this->settings['merchant_logo'];
	
		// Admin options
		add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );

		// Receipt
		add_action( 'woocommerce_receipt_' . $this->id, array( $this, 'initiate_deposit' ) );

		// Thank You
		// add_action( 'woocommerce_thankyou_' . $this->id, array( $this, 'success_message' ) );

		// API Callback
		add_action( 'woocommerce_api_' . $this->id, array( $this, 'check_transaction_status' ) );

    } // End Constructor


	/**
	 * Shows possible pending payment message
	 * 
	 * @since 1.0.0
	 * @access public
	 * @return void
	 */
	public function success_message() {
		echo '<div class="woocommerce-message">', __( 'Thanks for initiating the deposit. Your order will be processed once the payment is released.', 'astropay' ), '</div>';
	} // End success_message()


	/**
	 * Initialises Gateway Settings Form Fields
	 *
	 * @since 1.0.0
	 * @return void
	 */
	function init_form_fields() {
		$this->form_fields = array( 
			'enabled' => array( 
				'title' => __( 'Enable/Disable', 'astropay' ),
				'type' => 'checkbox',
				'label' => __( 'Enable AstroPay Payment Gateway', 'astropay' ),
				'default' => 'yes'
			 ),
			'title' => array( 
				'title' => __( 'Title', 'astropay' ),
				'type' => 'text',
				'description' => __( 'Payment method title that the customers see during the checkout.', 'astropay' ),
				'default' => __( 'Pay with AstroPay', 'astropay' )
			 ),
			'description' => array( 
				'title' => __( 'Description', 'astropay' ),
				'type' => 'textarea',
				'description' => __( 'Payment method description that the customers see during the checkout.', 'astropay' ),
				'default' => __( 'AstroPay offers multiple payment methods.', 'astropay' ),
			 ),
			'x_login' => array( 
				'title' => __( 'Merchant-Gateway-Api-Key', 'astropay' ),
				'type' => 'text',
				'description' => __( 'Merchant-Gateway-Api-Key is generated at the time of activation of your site and helps to uniquely identify you to AstroPay.', 'astropay' ),
				'default' => ''
			 ),
			'secret' => array( 
				'title' => __( 'Secret Key', 'astropay' ),
				'type' => 'password',
				'description' => __( 'Secret Key is generated at the time of activation of your site and helps to verify your identity to AstroPay.', 'astropay' ),
				'default' => ''
			 ),
			'testmode' => array( 
				'title' => __( 'Test Mode', 'astropay' ),
				'type' => 'checkbox',
				'label' => __( 'Enable AstroPay Test Environment. If checked, please make sure you use test keys above.', 'astropay' ),
				'default' => 'no'
			 ),
			'visual_info' => array(
				'title' => __( 'Visual Info (Optional)', 'astropay' ),
				'type' => 'title',
				'label' => __( 'Merchandising information to display on the actual payment screen.', 'astropay' ),
			 ),
			'merchant_name' => array( 
				'title' => __( 'Merchant Name', 'astropay' ),
				'type' => 'text',
				'description' => __( 'Merchant Name would appear on payment screen, and help users to know who is receiving the payment.', 'astropay' ),
				'default' => ''
			 ),
			 'merchant_logo' => array( 
				'title' => __( 'Merchant Logo (URL)', 'astropay' ),
				'type' => 'url',
				'description' => __( 'Merchant Logo image would appear on payment screen, and help users to see your identity.', 'astropay' ),
				'default' => ''
			 ),
		 );
	} // End init_form_fields()


	/**
	 * Displays Admin Panel Options
	 *
	 * @since 1.0.0
	 * @access public
	 * @return void
	 */
	public function admin_options() {
		echo '<h3>' . __( 'AstroPay Payment Gateway', 'astropay' ) . '</h3>';
		echo '<p>' . $this->method_description . '</p>';
		echo '<table class="form-table">';		
		$this->generate_settings_html(); // Generate the HTML For the settings form
		echo '</table>';
	} // End admin_options()


	/**
	 * Returns the description for the payment gateway
	 * 
	 * @since 1.0.0
	 * @access public
	 * @return string
	 */
	public function get_description() {
		return $this->description;
	} // End get_description()


	/**
	 * Returns the base URL depending on the mode
	 *
	 * @since 1.0.0
	 * @return string
	 */
	function get_base_url() {
		return $this->testmode ? $this->testurl : $this->liveurl;
	} // End get_base_url()


	/**
	 * Obtains payment methods list based on base country code
	 *
	 * @since 1.0.0
	 * @return array|WP_Error
	 */
	function get_payment_methods() {
		$country = WC()->customer->get_billing_country() ?: $this->country;

		$transient_id = 'apc_payment_methods_' . $country . '_' . (int) $this->testmode;

		$methods = get_transient( $transient_id );
		if ( $methods ) return $methods;

		$endpoint = $this->get_base_url() . 'merchant/v1/paymentMethods?country=' . $country;
		$methods = $this->get_data( $endpoint );

		if ( !is_wp_error( $methods ) ) {
			$methods = $methods['paymentMethods'] ?? [];
			if ( empty( $methods ) ) {
				return false;
			}
			set_transient( $transient_id, $methods, 3600 );
		}

		return $methods;
	} // End get_payment_methods()


    /**
	 * Add fields to pre-select method of payment
	 *
	 * @since 1.0.0
	 * @access public
	 * @return void
	 */
    public function payment_fields() {
		if ( $this->description ) {
			echo wpautop( wptexturize( $this->description ) );
		}

		$message = '';
		$methods = $this->get_payment_methods();
		if ( is_wp_error( $methods ) ) {
			$message = ( 'user_error' == $methods->get_error_code() ) ? $methods->get_error_message() : __( 'No payment method is available for your country.', 'astropay' );
			$methods = null;
		}
		?>
		<style>#payment_method_options>img{width:2em;max-height:unset;margin:.5em;float:unset;display:inline-block;vertical-align:middle}#payment_method_options>input[type="radio"]{vertical-align:middle}</style>
		<fieldset>
			<p class="form-row" id="payment_method_options">
				<?php
				if ( empty( $methods ) ) {
					if ( ! empty( $message ) ) {
						echo '<div class="woocommerce-error">' . $message . '</div>';
					}
				} else {
					echo '<label for="apcpm">', __( "Choose a Payment Method", 'astropay' ), '</label><br/>';
					foreach( $methods as $data ) {
						$code = $data['code'];
						echo '<input type="radio" name="apcpm" value="' . $code . '"/><img src="' . plugins_url( "assets/images/payment/{$code}_AP.png", __DIR__ ) . '" alt="' . $code . '"/>' . self::PAYMENT_METHODS[$code] . '<br/>';
					}
				}?>
			</p>
			<div class="clear"></div>
		</fieldset>
		<?php
    } // End payment_fields()


	/**
     * Return the current transaction currency
	 * - Supports WOOCS currency switcher
	 * 
	 * @since 1.0.0
	 * @return string
	 */
    function get_currency() {
        if ( class_exists( 'WOOCS' ) ) {
            global $WOOCS;
            $currency = strtoupper( $WOOCS->storage->get_val( 'woocs_current_currency' ) );
        } else {
            $currency = get_woocommerce_currency();
        }
        return $currency;
	} // End get_currency()
	

	/**
	 * Check the validity of data recived and update the status of order
	 *
	 * @since 1.0.0
	 * @return void
	 */
	function check_transaction_status() {
		global $woocommerce;

		// Obtain the callback request type
		$this->log_message( 'check_transaction_status triggered with params', $_REQUEST );
		$trans_type = $this->get_get_param( 'type' );

		if ( $trans_type == 'deposit' ) {
			$this->check_deposit_status();
		} else {
			$this->check_refund_status();
		}

		exit;
	} // End check_transaction_status()


	/**
     * Returns the array containing user data from the order
	 * 
	 * @since 1.0.0
	 * @param WP_Order	$order	Order from which user data needs to be extracted
	 * @return array
	 */
	function get_user_data( $order ) {
		$phone = $order->get_billing_phone();
		$country = $order->get_billing_country();
		$country_phone = self::sanitize_phone_number( $phone, $country );
		
		return [
			'merchant_user_id' => $order->get_customer_id(),
			'email' => $order->get_billing_email(),
			'phone' => $country_phone,
			'first_name' => $order->get_billing_first_name(),
			'last_name' => $order->get_billing_last_name(),
			'address' => [
				'line1' => $order->get_billing_address_1(),
				'line2' => $order->get_billing_address_2(),
				'city' => $order->get_billing_city(),
				'province' => $order->get_billing_state(),
				'country' => $country,
				'zip' => $order->get_billing_postcode()
			]
		];
	} // End get_user_data()


	/**
	 * Process the payment and return the result.
	 *
	 * @since 1.0.0
	 * @access public
	 * @param int	order_id	Order id for which payment is being made
	 * @return array
	 */
	public function process_payment( $order_id ) {
		$order = new WC_Order( $order_id );

		$apcpm = $this->get_post_param( 'apcpm' ) ?? '';

		return array( 
			'result' 	=> 'success',
			'redirect'	=> add_query_arg( 'order', $order_id, add_query_arg( 'key', $order->get_order_key(), add_query_arg( 'apcpm', $apcpm, $order->get_checkout_payment_url( true ) ) ) )
		 );
	} // End process_payment()


	/**
	 * Initiate deposit transaction for the order
	 * 
	 * @since 1.0.0
	 * @param int	order_id	Order id for which payment is being made
	 * @return void
	 */
	function initiate_deposit( $order_id ) {
		echo '<style>.spinner{width:10em;height:10em;margin:2em auto;position:relative;font-size:1em;text-indent:-9999em;border:1em solid #0003;border-radius:50%;border-top-color:#000F;-webkit-transform:translateZ(0);-ms-transform:translateZ(0);transform:translateZ(0);-webkit-animation:spin 1s infinite linear;animation:spin 1s infinite linear}@-webkit-keyframes spin{0%{-webkit-transform:rotate(0deg);transform:rotate(0deg)}100%{-webkit-transform:rotate(360deg);transform:rotate(360deg)}}@keyframes spin{0%{-webkit-transform:rotate(0deg);transform:rotate(0deg)}100%{-webkit-transform:rotate(360deg);transform:rotate(360deg)}}</style>';
		echo '<p>' . __( 'Thank you for your order! Please wait while we redirect you to Astropay for payment.', 'astropay' ) . '</p>';
		echo '<div class="spinner">Redirecting...</div>';

		$order = wc_get_order( $order_id );

		$deposit_id = wp_generate_uuid4();
		$redirect_key = wp_generate_uuid4();
		$callback_key = wp_generate_uuid4();

		update_post_meta( $order_id, '_apc_order_keys', [
			'callback' => $callback_key,
			'redirect' => $redirect_key,
		] );

		$order_return_url = add_query_arg( ['wc-api' => $this->id, 'id' => $order_id, 'type' => 'deposit' ], trailingslashit( get_home_url() ) );

		$data = [
			'amount' => round( $order->get_total(), 2 ),
			'currency' => $order->get_currency(),
			'country' => $this->country,
			'merchant_deposit_id' => $deposit_id,
			'callback_url' => add_query_arg( 'key', $callback_key, $order_return_url ),
			'redirect_url' => add_query_arg( ['redirect' => 1, 'key' => $redirect_key], $order_return_url ),
			'user' => $this->get_user_data( $order ),
			'product' => [
				'mcc' => 7995,
				'category' => 'eCommerce',
				'merchant_code' => 'eBusiness',
				'description' => 'eCommerce Business'
			],
		];

		$method = $this->get_get_param( 'apcpm' );
		if ( !empty( $method ) ) {
			$data['payment_method_code'] = $method;
			update_post_meta( $order_id, '_apc_payment_method', $method );
		}

		if ( !empty( $this->merchant_name ) || !empty( $this->merchant_logo ) ) {
			$data['visual_info'] = [];
			if ( !empty( $this->merchant_name ) ) {
				$data['visual_info']['merchant_name'] = $this->merchant_name;
			}
			if ( !empty( $this->merchant_logo ) ) {
				$data['visual_info']['merchant_logo'] = $this->merchant_logo;
			}
		}

		$endpoint = $this->get_base_url() . 'merchant/v1/deposit/init';
		$result = $this->post_data( $endpoint, $data );
		
		if ( is_wp_error( $result ) ) {
			$error = __( $result->get_error_message(), 'astropay' );
			$this->log_message( "Deposit Initiation Error: $error" );
			echo '<script>window.location = "', $order->get_checkout_payment_url(), '";</script>';
			wc_add_notice( $error, 'error' );
		} else {
			$this->log_message( "Deposit Initiation Result: ", $result );
			$deposit_id = $result['deposit_external_id'];
			update_post_meta( $order_id, '_apc_deposit_id', $deposit_id );
			echo '<script>window.location = "', $result['url'], '";</script>';
		}
	} // End initiate_deposit()


	/**
	 * Check the validity of data recived and update the status of order
	 *
	 * @since 1.0.0
	 * @return void
	 */
	function check_deposit_status() {
		// Obtain the query parameters
		$order_id = $this->get_get_param( 'id' );
		$order_key = $this->get_get_param( 'key' );
		$redirect = $this->get_get_param( 'redirect' );

		// Verify the unique key
		$order_keys = get_post_meta( $order_id, '_apc_order_keys', true );
		$unique_key = $redirect && isset( $order_keys['redirect'] ) ? $order_keys['redirect'] : $order_keys['callback'];
		if ( $unique_key != $order_key ) {
			$this->log_message( 'Ignoring the callback due to unique key mismatch' );
			die( 'Incorrect key!' );
		}

		// Obtain the deposit status data as per the nature of request
		$deposit_id = get_post_meta( $order_id, '_apc_deposit_id', true );
		$order = wc_get_order( $order_id );
		if ( $redirect ) {	//immediate callback
			$data = $this->query_deposit_status( $deposit_id );
			if ( is_wp_error( $data ) ) {	// handle the error immediately
				$error = $data->get_error_message();
				wc_add_notice( __( $error, 'astropay'), 'error' );
				wp_redirect( $order->get_checkout_payment_url() );
				return;
			}
		} else {		// later callback
			// Capture JSON Input
			$input = file_get_contents( 'php://input' );
			$data = json_decode( $input, true );
			if ( !$data ) {	// Capture URL Encode Input
				parse_str( $input, $data );
			}
			if ( $deposit_id != $data['deposit_external_id'] ) {
				$this->log_message( "Invalid callback for order ID $order_id having deposit id $deposit_id", $data );
				die( 'Incorrect Deposit ID' );
			}
		}

		// Update the deposit status on the order
		$this->log_message( 'transaction data', $data );
		$status = $data['status'];
		$result = $this->update_order_deposit_status( $order, $deposit_id, $status );

		// Obtain the updated order status, and redirect the user if needed
		$order_status = $order->get_status();
		$this->log_message( "Updated order status to $order_status for Order ID $order_id" );
		if ( $redirect ) {
			if ( !in_array( $order_status, ['pending', 'failed'] ) ) {
				if ( $woocommerce->cart ) $woocommerce->cart->empty_cart();
				if ( 'failed' == $order_status ) {
					wp_redirect( $order->get_cancel_order_url() );
				} else {
					wp_redirect( $order->get_checkout_order_received_url() );
				}
			} else {
				wp_redirect( $order->get_checkout_payment_url() );
			}
		} else {
			die( $result );
		}
	} // End check_deposit_status()


	/**
	 * Queries the status of a transaction
	 * 
	 * @since 1.0.0
	 * @param int	deposit_id	Id of the deposit for which status is queried
	 * @return array|WP_Error
	 */
	function query_deposit_status( $deposit_id ) {
		$endpoint = $this->get_base_url() . 'merchant/v1/deposit/' . $deposit_id . '/status';
		$data = $this->get_data( $endpoint );

		if ( is_wp_error( $data ) ) {
			$this->log_message( $data->get_error_message() );
		}

		return $data;
	} // End query_deposit_status()


	/**
	 * Update the deposit status on order
	 * 
	 * @since 1.0.0
	 * @param WC_Order	order		Order whose deposit status needs to be updated
	 * @param int		deposit_id	Id of the deposit for which status is updated
	 * @param string	status		Status of transaction as obtained from astropay api
	 * @return boolean
	 */
	function update_order_deposit_status( $order, $deposit_id, $status ) {
		$order_id = $order->get_id();
		$old_status = $order->get_status();

		$note = __( 'AstroPay deposit id', 'astropay' ) . ': ' . $deposit_id . ' - ';

		switch( $status ) {
			case 'PENDING':
				$note .= __( 'The deposit is created but the customer is yet to complete the payment flow.', 'astropay' );
				$new_status = 'pending';
			break;
			case 'CANCELLED':
				$note .= __( 'The customer cancelled the payment for the deposit.', 'astropay' );
				$new_status = 'failed';
			break;
			case 'APPROVED':
				$note .= __( 'The deposit has been completed and the money was credited to your account.', 'astropay' );
				$order->add_order_note( $note );
				$order->payment_complete();
			break;
			default:
				$note .= __( 'Ignored unhandled status: ', 'astropay' ) . $status;
		}

		if ( $new_status && $new_status != $old_status ) {
			$order->add_order_note( $note );
			$order->update_status( $new_status );
		}
		$this->log_message( "Order ID: $order_id - $note" );
		return true;
	} // End update_order_deposit_status()


	/**
	 * Process the refund and return the result.
	 *
	 * @since 1.0.0
	 * @access public
	 * @param int		$order_id	Order ID.
	 * @param mixed		$amount		The amount to refund
	 * @param string	$reason		The reason for refund
	 * @return boolean
	 */
	public function process_refund( $order_id, $amount = NULL, $reason = '' ) {
		if ( empty( $order_id ) ) return false;
		
		if ( empty( $amount ) ) {
			$this->log_message( "Skipping ZERO value refund for Order ID $order_id ");
			return false;
		}
		
		$order_return_url = add_query_arg( ['wc-api' => $this->id, 'id' => $order_id, 'type' => 'refund' ], trailingslashit( get_home_url() ) );

		$cashout_id = wp_generate_uuid4();
		$callback_key = wp_generate_uuid4();

		$order = wc_get_order( $order_id );
		$data = [
			'amount' => round( $amount, 2 ),
			'currency' => $order->get_currency(),
			'country' => $this->country,
			'merchant_cashout_id' => $cashout_id,
			'callback_url' => add_query_arg( 'key', $callback_key, $order_return_url ),
			'user' => $this->get_user_data( $order ),
		];

		$endpoint = $this->get_base_url() . 'merchant/v1/cashout';
		$result = $this->post_data( $endpoint, $data );

		if ( is_wp_error( $result ) ) {
			$error = __( $result->get_error_message(), 'astropay' );
			$order->add_order_note( "Refund Failure - $error" );
			$this->log_message( "Refund Error: $error" );
			return false;
		} else {
			$status = $result['status'];
			// $cashout_id = $result['merchant_cashout_id'];
			update_post_meta( $order_id, '_apc_refund_id', $cashout_id );
			update_post_meta( $order_id, '_apc_refund_key', $callback_key );
			return $this->update_order_refund_status( $order, $cashout_id, $status );
		}
	} // End process_refund()


	/**
	 * Check the validity of data recived and update the status of order
	 *
	 * @since 1.0.0
	 * @return void
	 */
	function check_refund_status() {
		// Obtain the query parameters
		$order_id = $this->get_get_param( 'id' );
		$order_key = $this->get_get_param( 'key' );

		// Verify the unique key
		$refund_key = get_post_meta( $order_id, '_apc_refund_key', true );
		if ( $refund_key != $order_key ) {
			$this->log_message( 'Ignoring the callback due to unique key mismatch' );
			die( 'Incorrect key!' );
		}

		// Obtain the refund status data
		$input = file_get_contents( 'php://input' ); // Capture JSON Input
		$data = json_decode( $input, true );
		if ( !$data ) {	// Capture URL Encode Input
			parse_str( $input, $data );
		}
		$refund_id = get_post_meta( $order_id, '_apc_refund_id', true );
		if ( $refund_id != $data['merchant_cashout_id'] ) {
			$this->log_message( "Invalid callback for order ID $order_id having refund id $refund_id", $data );
			die( 'Incorrect Cashout ID' );
		}

		// Update the refund status on the order
		$this->log_message( 'transaction data', $data );
		$status = $data['status'];
		$order = wc_get_order( $order_id );
		$result = $this->update_order_refund_status( $order, $refund_id, $status );
		die( $result );
	} // End check_refund_status()


	/**
	 * Update the refund status on order
	 * 
	 * @since 1.0.0
	 * @param WC_Order	order		Order whose refund status needs to be updated
	 * @param int		refund_id	Id of the refund for which status is updated
	 * @param string	status		Status of transaction as obtained from astropay api
	 * @return boolean
	 */
	function update_order_refund_status( $order, $refund_id, $status ) {
		$order_id = $order->get_id();

		$note = __( 'AstroPay refund id', 'astropay' ) . ': ' . $refund_id . ' - ';

		switch( $status ) {
			case 'PENDING':
				$note .= __( 'The refund is created but is yet to be processed.', 'astropay' );
			break;
			case 'CANCELLED':
				$note .= __( 'The merchant cancelled the payment for the refund.', 'astropay' );
				$order->update_status( 'cancelled' );
			break;
			case 'APPROVED':
				$note .= __( 'The refund has been completed and the money was creditd to customer card.', 'astropay' );
			break;
			default:
				$note .= __( 'Ignored unhandled status: ', 'astropay' ) . $status;
		}

		$this->log_message( $note );
		$order->add_order_note( $note );

		return 'CANCELLED' != $status;
	} // End update_order_deposit_status()


	/**
	 * Creates the headers for authorization with Direct24
	 * 
	 * @since 1.0.0
	 * @param string	body	Post payload to be used for calculating hash
	 * @return array
	 */
	function create_headers( $body='' ) {
		return [
			'Merchant-Gateway-Api-Key' => $this->x_login,
			'Signature' => hash_hmac( 'sha256', $body, $this->secret ),
			'Signature-Algorithm' => 'HMACSHA256',
			'Content-Type' => 'application/json'
		];
	} // End create_headers()


	/**
	 * Parse HTTP response and returns relevant data
	 * 
	 * @since 1.0.0
	 * @param array|WP_Error	response	Response received from HTTP request
	 * @return array|WP_Error
	 */
	function process_http_response( $response ) {
		$this->log_message( "Processing Response", $response );

		if ( is_wp_error( $response ) ) return $response;

		if ( !is_array( $response ) ) return new WP_Error( 'response', $response );

		$body = $response['body'];
		if ( is_string( $body ) ) {
			$body = json_decode( $body, true );
		}

		$code = $response['response']['code'] ?? 0;
		if ( $code >= 400 ) return new WP_Error( 'api_error', self::ERRORS[ $code ] ?? $response['response']['message'] );

		if ( 200 < $code || $code >= 300 ) return new WP_Error( 'api_error', $body['description'] ?? $response['response']['message'] );

		if ( !empty( $body['error'] ) ) return new WP_Error( 'api_error', $body['error'] );

		return $body;
	} // End process_http_response()


	/**
	 * Gets the remote data
	 * 
	 * @since 1.0.0
	 * @param string	url			Request url
	 * @return array|WP_Error
	 */
	function get_data( $url ) {
		$options =  array( 
			'headers' => $this->create_headers(),
			'timeout' => 120,
			'sslverify' => false,
		 );

		$this->log_message( "Getting $url with options", $options );

		$response = wp_remote_get( $url, $options );

		return $this->process_http_response( $response );
	} // End get_data()


	/**
	 * Puts the data to remote URL and returns the response
	 * 
	 * @since 1.0.0
	 * @param string	url 		Request url
	 * @param array		data 		Data to be posted to the URL
	 * @return array|WP_Error
	 */
	function post_data( $url, $data ) {
		// $body = utf8_encode( json_encode( $data ) );
		$body = json_encode( $data );

		$headers = $this->create_headers( $body );

		$options = array( 
			'headers' => $headers,
			'timeout' => 120,
			'sslverify' => false,
			'body' => $body,
		 );

		$this->log_message( "Posting to $url with headers", $headers );
		$this->log_message( "Data body", $body );

		$response = wp_remote_post( $url, $options );

		return $this->process_http_response( $response );
	} // End post_data()


	/**
	 * Get post data if set
	 *
	 * @since 1.0.0
	 * @param string	name	Name of the variable to be obtained
	 * @return mixed
	 */
	function get_post_param( $name ) {
		return isset( $_POST[$name] ) ? sanitize_text_field( $_POST[$name] ) : null;
	} // End get_post_param()


	/**
	 * Get get data if set
	 *
	 * @since 1.0.0
	 * @param string	name	Name of the variable to be obtained
	 * @return mixed
	 */
	function get_get_param( $name ) {
		return isset( $_GET[$name] ) ? sanitize_text_field( $_GET[$name] ) : null;
	} // End get_get_param()


	/**
	 * Log a message to log file
	 * 
	 * @since 1.0.0
	 * @param string	message	Message to be logged
	 * @param mixed		data	Additional data to be logged as json encoded string
	 * @param string	sep		Separator to be used for appending data to message
	 * @return void
	 */
	function log_message( $message, $data=null, $sep=': ' ) {
		if ( !empty( $data ) ) {
			$message .= $sep . ( is_wp_error( $data ) ? $data->get_error_message() : ( is_string( $data ) ? $data : json_encode( $data ) ) );
		}
		$this->logger->add( $this->id, $message );
	} // End log_message()


	/**
	 * Adds country calling code to the phone number if not added already
	 * 
	 * @since 1.0.0
	 * @param string	number	Phone number to be sanitized
	 * @param string	country	ISO country code
	 * @return string
	 */
	public static function sanitize_phone_number( $number, $country ) {
		//Convert phone number to numeric value
		$phone = str_replace( array( '+', '-' ), '', filter_var( $number, FILTER_SANITIZE_NUMBER_INT ) );
		$phone = ltrim( $phone, '0' );

		//Obtain international country calling code for the given country
		$intl_prefix = WC()->countries->get_country_calling_code( $country );
        $intl_prefix = is_array( $intl_prefix ) ? $intl_prefix[0] : $intl_prefix;

		//Prefix to phone if not already added
		preg_match( "/(\d{1,4})[0-9.\- ]+/", $phone, $prefix );
		if ( strpos( $prefix[1], $intl_prefix ) !== 0 ) {
			$phone = $intl_prefix . $phone;
		}
		
		//Return the prefixed phone number
		return $phone;	
	} // End sanitize_phone_number()

}; //  End Class
